#=======================================================================
#	$Id: PHP.pm,v 1.3 2005/09/09 11:20:48 pythontech Exp $
#	Store session data in PHP-compatible session files.
#	Copyright (C) 2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#
#	Ideas pinched from PHP-Session
#	  !name|
#	  name|value;
#=======================================================================
package PythonTech::SessionDB::PHP;
use strict;

sub new {
    my $class = shift;
    my $self = {@_};
    $self->{'sessionDir'} ||= '/tmp';
    bless $self, $class;
    return $self;
}

#-----------------------------------------------------------------------
#	Find or create a session
#-----------------------------------------------------------------------
sub find_session {
    my($self, $query) = @_;
    my $tsessid = $query->cookie('PHPSESSID');
#    print STDERR "PHPSESSID=$tsessid\n";
    if (my($claimed_id) = $tsessid =~ m!^([0-9a-f]{32})$!) {
	my $sessfile = "$self->{'sessionDir'}/sess_$claimed_id";
#	print STDERR "sessfile=$sessfile\n";
	if (-f $sessfile && -r $sessfile && -w $sessfile) {
#	    print STDERR "file -f,-r,-w\n";
	    my $session = {
		db => $self,
		id => $claimed_id,
		filename => $sessfile,
	    };
	    open(PHPSESS,'<',$sessfile)
		or die "Cannot open $sessfile: $!";
	    my $text = join('',<PHPSESS>);
	    close(PHPSESS);
#	    print STDERR "text=$text\n";
	    while ($text =~ m!\G([a-z]\w*)\|!igc) {
		my $key = $1;
		my $value = &_decode(\$text);
#		print STDERR "key $key=$value\n";
		$session->{".$key"} = $value;
	    }
	    bless $session, ref($self)."::Session";
#	    print STDERR "find $session all:",(map{" $_=$session->{$_}"} keys %$session),"\n";
	    return $session;
	}
    }
    # Else - no cookie, or invalid, or missing file, or unwritable
    for (my $try=1; $try < 5; $try++) {
	eval 'use Digest::MD5 qw(md5_hex)';
	my $id = md5_hex((rand 2147483647 ^ $$) .
			 (rand 2147483647 ^ time) .
			 rand 2147483647 .
			 rand 2147483647);
	my $sessfile = "$self->{'sessionDir'}/sess_$id";
	if (! -f $sessfile) {
	    # FIXME race condition - lock or sysopen with options...?
	    my $oldumask = umask(077);
	    if (open(PHPSESS,'>',$sessfile)) {
		umask($oldumask);
		close(PHPSESS);
		my $session = {
		    db => $self,
		    id => $id,
		    filename => $sessfile,
		};
		bless $session, ref($self)."::Session";
		$query->response->set_cookie(-name => 'PHPSESSID',
					     -value => $id);
		return $session; # Success

	    } else {
		# Failed to open - try another id
		umask($oldumask);
	    }
	}
    }
    die "Cannot allocate session token after tries";
}

sub _encode {
    my($value) = @_;
    if (! defined $value) {
	return "N;";
    } elsif (! ref($value)) {
	if ($value =~ /^-?\d+$/) {
	    return "i:$value;";
	} elsif ($value =~ /^-?\d+(\.\d+)?([eE]-?\d+)?$/) {
	    return "d:" . uc($value) . ";";
	} else {
	    return "s:" . length($value) . ":\"$value\";";
	}
    } elsif (ref($value) eq 'ARRAY') {
	my $alen = scalar @$value;
	my $a = "a:$alen:{";
	for (my $i=0; $i < $alen; $i++) {
	    $a .= "i:$i;" . _encode($value->[$i]);
	}
	$a .= "}";
	return $a;
    } elsif (ref($value) eq 'HASH') {
	my $alen = scalar keys %$value;
	my $a = "a:$alen:{";
	while (my($key,$val) = each %$value) {
	    $a .= _encode($key) . _encode($val);
	}
	$a .= "}";
	return $a;
    } else {
	die "Cannot encode ".ref($value)."\n";
    }
}

sub _decode {
    my($ptext) = @_;
    if ($$ptext =~ /\Gi:(-?\d+);/gc) {
	# integer e.g. i:123;
	return $1;
    } elsif ($$ptext =~ /\Gd:(-?\d+(?:\.\d+)?(?:E[-+]\d+)?);/gc) {
	# double e.g. d:2.156E+18;
	return $1;
    } elsif ($$ptext =~ /\GN;/gc) {
	# Null e.g. N;
	return undef;
    } elsif ($$ptext =~ /\Gb:([01]);/gc) {
	# boolean e.g. b:1;
	return $1;
    } elsif ($$ptext =~ /\Gs:(\d+):/gc) {
	# string e.g. s:11:"Hello world";
	my $slen = $1;
#	print STDERR "len=$slen string=",substr($$ptext,pos($$ptext),10),"\n";
	$$ptext =~ /\G\"(.{$slen})\";/sgc
	    or die "String mismatch\n";
	return $1;
    } elsif ($$ptext =~ /\Ga:(\d+):\{/gc) {
	# array e.g. a:2:{s:3:"foo";i:123;s:3:"bar";s:4:"quux";}
	my $alen = $1;
	my %a;
	my @a;
	for (my $i=0; $i<$alen; $i++) {
	    my $key = _decode($ptext);
	    my $value = _decode($ptext);
	    $a{$key} = $value;
	    push @a, $value if $key eq $i;
	}
	if (@a == $alen) {
	    # Keys were 0,1,2,...
	    return \@a;
	} else {
	    return \%a;
	}
    } elsif ($$ptext =~ /\GO:(\d+):/gc) {
	# Object e.g. O:3:"Foo":2:{s:1:"a";i:1;s:1:"b";i:2;}
	my $classlen = $1;
	$$ptext =~ /\G\"([.\n]{$classlen})\":(\d+):/gc
	    or die "Error in class definition\n";
	my($class,$olen) = ($1,$2);
	my %o;
	for (my $i=0; $i<$olen; $i++) {
	    my $key = _decode($ptext);
	    my $value = _decode($ptext);
	    $o{$key} = $value;
	}
	$o{'_class'} = $class;
	return bless \%o, 'PythonTech::SesssionDB::PHP::Object';
    } else {
	die "Invalid PHP data at ".substr($$ptext,pos($$ptext),10)."\n";
    }
}

#=======================================================================
#	Session object
#=======================================================================
package PythonTech::SessionDB::PHP::Session;

sub get {
    my($self, $key) = @_;
#    print STDERR "get $self key=$key all=",(map{"$_=$self->{$_}"} keys %$self),"\n";
    return $self->{".$key"};
}

sub set {
    my($self, $key,$value) = @_;
    $self->{".$key"} = $value;	# FIXME deepcopy?
    $self->{'modified'} = 1;
}

sub login {
    my($self, $username) = @_;
    $self->set("userName", $username);
}

sub clear {
    my($self, $keylist) = @_;
    foreach my $key (@$keylist) {
	delete $self->{".$key"};
    }
    $self->{'modified'} = 1;
}

#-----------------------------------------------------------------------
#	Write changes to session file
#-----------------------------------------------------------------------
sub flush {
    my($self) = @_;
    if ($self->{'modified'}) {
	my $newfile = "$self->{'filename'}.$$";
	open(PHPSESS,'>',$newfile)
	    or die "Cannot open $newfile: $!";
	while (my($key,$value) = each %{$self}) {
	    if (substr($key,0,1) eq '.') {
		print PHPSESS (substr($key,1),"|",
			       &PythonTech::SessionDB::PHP::_encode($value));
	    }
	}
	close(PHPSESS)
	    or die "Error closing $newfile: $!";
	rename($newfile, $self->{'filename'})
	    or die "Error renaming $newfile: $!";
    }
    $self->{'modified'} = undef;
}

1;
